# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

import os
import sys

project = 'ez-a-sync'
copyright = '2024, BobTheBuidler'
author = 'BobTheBuidler'

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    'sphinx.ext.autodoc',
    'sphinx.ext.napoleon',
]

templates_path = ['_templates']
exclude_patterns = ['_build', 'Thumbs.db', '.DS_Store']



# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = 'alabaster'
html_static_path = ['_static']

autodoc_default_options = {
    'private-members': True,
    # hide private methods that aren't relevant to us here
    #'exclude-members': f'{_pony_private_members},_abc_impl,_prune_running,_ConstantSingletonMeta__instances,_ConstantSingletonMeta__lock,_do_math,_coros',
}
automodule_generate_module_stub = True

sys.path.insert(0, os.path.abspath('./a_sync'))




import asyncio
import sys

from a_sync._typing import *
from a_sync.primitives._debug import _DebugDaemonMixin


class Event(asyncio.Event, _DebugDaemonMixin):
    """asyncio.Event but with some additional debug logging to help detect deadlocks."""
    _value: bool
    _loop: asyncio.AbstractEventLoop
    _waiters: Deque["asyncio.Future[None]"]
    
    def __init__(self, name: str = "", debug_daemon_interval: int = 300, *, loop: Optional[asyncio.AbstractEventLoop] = None):
        if sys.version_info >= (3, 10):
            super().__init__()
        else:
            super().__init__(loop=loop)
        self._name = name
        self._loop = self._loop or asyncio.get_event_loop()
        self._debug_daemon_interval = debug_daemon_interval
    def __repr__(self) -> str:
        label = f'name={self._name}' if self._name else 'object'
        status = 'set' if self._value else 'unset'
        if self._waiters:
            status += f', waiters:{len(self._waiters)}'
        return f"<{self.__class__.__module__}.{self.__class__.__name__} {label} at {hex(id(self))} [{status}]>"
    async def wait(self) -> Literal[True]:
        if self.is_set():
            return True
        self._ensure_debug_daemon()
        return await super().wait()
    async def _debug_daemon(self) -> None:
        while not self.is_set():
            await asyncio.sleep(self._debug_daemon_interval)
            if not self.is_set():
                self.logger.debug("Waiting for %s", self)



import asyncio
from collections import defaultdict
from time import time
from typing import DefaultDict, Iterable, Optional

from a_sync.primitives._debug import _DebugDaemonMixin
from a_sync.primitives.locks.event import Event


class CounterLock(_DebugDaemonMixin):
    """
    A asyncio primative that blocks until the internal counter has reached a specific value.
    
    counter = CounterLock()
    A coroutine can now `await counter.wait_for(3)` and it will block until the internal counter >= 3.
    Now if some other task executes `counter.value = 5` or `counter.set(5)`, the first coroutine will unblock as 5 >= 3.
    
    The internal counter can only increase.
    """
    def __init__(self, start_value: int = 0, name: Optional[str] = None):
        self._name = name
        self._value = start_value
        self._events: DefaultDict[int, Event] = defaultdict(Event)
        self.is_ready = lambda v: self._value >= v
        
    async def wait_for(self, value: int) -> bool:
        if not self.is_ready(value):
            self._ensure_debug_daemon()
            await self._events[value].wait()
        return True
    
    def set(self, value: int) -> None:
        self.value = value
    
    def __repr__(self) -> str:
        waiters = {v: len(self._events[v]._waiters) for v in sorted(self._events)}
        return f"<CounterLock name={self._name} value={self._value} waiters={waiters}>"
        
    @property
    def value(self) -> int:
        return self._value
    
    @value.setter
    def value(self, value: int) -> None:
        if value > self._value:
            self._value = value
            ready = [self._events.pop(key) for key in list(self._events.keys()) if key <= self._value]
            for event in ready:
                event.set()
        elif value < self._value:
            raise ValueError("You cannot decrease the value.")
    
    async def _debug_daemon(self) -> None:
        start = time()
        while self._events:
            self.logger.debug("%s is still locked after %sm", self, round(time() - start / 60, 2))
            await asyncio.sleep(300)

class CounterLockCluster:
    """
    An asyncio primitive that represents 2 or more CounterLock objects.
    
    `wait_for(i)` will block until the value of all CounterLock objects is >= i.
    """
    def __init__(self, counter_locks: Iterable[CounterLock]) -> None:
        self.locks = list(counter_locks)
    
    async def wait_for(self, value: int) -> bool:
        await asyncio.gather(*[counter_lock.wait_for(value) for counter_lock in self.locks])
        return True
    




import pytest

from a_sync.iter import (ASyncIterable, ASyncIterator, ASyncWrappedIterable,
                         ASyncWrappedIterator)


async def async_gen():
    yield 0
    yield 1
    yield 2

def test_iterable_wrap():
    assert isinstance(ASyncIterable.wrap(async_gen()), ASyncWrappedIterable)

def test_iterator_wrap():
    assert isinstance(ASyncIterator.wrap(async_gen()), ASyncWrappedIterator)

def test_iterable_sync():
    assert [i for i in ASyncIterable.wrap(async_gen())] == [0, 1, 2]

@pytest.mark.asyncio_cooperative
async def test_iterable_async():
    assert [i async for i in ASyncIterable.wrap(async_gen())] == [0, 1, 2]

def test_iterator_sync():
    iterator = ASyncIterator.wrap(async_gen())
    for i in range(4):
        if i < 3:
            assert next(iterator) == i
        else:
            with pytest.raises(StopIteration):
                next(iterator)

@pytest.mark.asyncio_cooperative   
async def test_iterator_async():
    iterator = ASyncIterator.wrap(async_gen())
    for i in range(4):
        if i < 3:
            assert await iterator.__anext__() == i
        else:
            with pytest.raises(StopAsyncIteration):
                await iterator.__anext__()




import asyncio
import logging

from a_sync._typing import *
from a_sync import exceptions
from a_sync.iter import ASyncIterable
from a_sync.utils.as_completed import as_completed
from a_sync.utils.gather import gather
from a_sync.utils.iterators import as_yielded, exhaust_iterator


logger = logging.getLogger(__name__)

def create_task(coro: Awaitable[T], *, name: Optional[str] = None, skip_gc_until_done: bool = False) -> "asyncio.Task[T]":
    """A wrapper over `asyncio.create_task` which will work with any `Awaitable` object, not just `Coroutine` objects"""
    if not asyncio.iscoroutine(coro):
        coro = __await(coro)
    task = asyncio.create_task(coro, name=name)
    if skip_gc_until_done:
        __persist(asyncio.create_task(__persisted_task_exc_wrap(task)))
    return task


MappingFn = Callable[Concatenate[K, P], Awaitable[V]]

class TaskMapping(ASyncIterable[Tuple[K, V]], DefaultDict[K, "asyncio.Task[V]"]):
    def __init__(self, coro_fn: MappingFn[K, P, V] = None, *iterables: AnyIterable[K], name: str = '', **coro_fn_kwargs: P.kwargs) -> None:
        self._coro_fn = coro_fn
        self._coro_fn_kwargs = coro_fn_kwargs
        self._name = name
        self._loader: Optional["asyncio.Task[None]"]
        if iterables:
            self._loader = create_task(exhaust_iterator(self._tasks_for_iterables(*iterables)))
        else:
            self._loader = None
    def __setitem__(self, item: Any, value: Any) -> None:
        raise NotImplementedError("You cannot manually set items in a TaskMapping")
    def __getitem__(self, item: K) -> "asyncio.Task[V]":
        try:
            return super().__getitem__(item)
        except KeyError:
            task = create_task(
                coro=self._coro_fn(item, **self._coro_fn_kwargs),
                name=f"{self._name}[{item}]" if self._name else f"{item}",
            )
            super().__setitem__(item, task)
            return task
    def __await__(self) -> Generator[Any, None, Dict[K, V]]:
        """await all tasks and returns a mapping with the results for each key"""
        return self._await().__await__()
    async def __aiter__(self) -> AsyncIterator[Tuple[K, V]]:
        """aiterate thru all key-task pairs, yielding the key-result pair as each task completes"""
        yielded = set()
        # if you inited the TaskMapping with some iterators, we will load those
        if self._loader:
            while not self._loader.done():
                async for key, value in self.yield_completed(pop=False):
                    if key not in yielded:
                        yield _yield(key, value, "both")
                        yielded.add(key)
                await asyncio.sleep(0)
            # loader is already done by this point, but we need to check for exceptions
            await self._loader
        elif not self:
            # if you didn't init the TaskMapping with iterators and you didn't start any tasks manually, we should fail
            raise exceptions.MappingIsEmptyError
        # if there are any tasks that still need to complete, we need to await them and yield them
        if self:
            async for key, value in as_completed(self, aiter=True):
                if key not in yielded:
                    yield _yield(key, value, "both")
    #def keys(self) -> KeysView[K]:
    #    if self._loader and not self._loader.done():
    #        raise RuntimeError("the loader needs time to complete. bob will figure out a way to make this not impact sync users")
    #    return super().keys()
    async def map(self, *iterables: AnyIterable[K], pop: bool = True, yields: Literal['keys', 'both'] = 'both') -> AsyncIterator[Tuple[K, V]]:
        if self:
            raise exceptions.MappingNotEmptyError
        else:
            logger.info(self)
        async for _ in self._tasks_for_iterables(*iterables):
            async for key, value in self.yield_completed(pop=pop):
                yield _yield(key, value, yields)
        async for key, value in as_completed(self, aiter=True):
            if pop:
                self.pop(key)
            yield _yield(key, value, yields)
    async def yield_completed(self, pop: bool = True) -> AsyncIterator[Tuple[K, V]]:
        for k, task in dict(self).items():
            if task.done():
                if pop:
                    task = self.pop(k)
                yield k, await task
    async def _await(self) -> Dict[K, V]:
        if self._loader:
            await self._loader
        if not self:
            raise exceptions.MappingIsEmptyError
        return await gather(self)
    async def _tasks_for_iterables(self, *iterables) -> AsyncIterator["asyncio.Task[V]"]:
        async for key in as_yielded(*[_yield_keys(iterable) for iterable in iterables]):  # type: ignore [attr-defined]
            yield self[key]  # ensure task is running
        

__persisted_tasks: Set["asyncio.Task[Any]"] = set()

async def __await(awaitable: Awaitable[T]) -> T:
    return await awaitable
    
def __persist(task: "asyncio.Task[Any]") -> None:
    __persisted_tasks.add(task)
    __prune_persisted_tasks()

def __prune_persisted_tasks():
    for task in tuple(__persisted_tasks):
        if task.done():
            if (e := task.exception()) and not isinstance(e, exceptions.PersistedTaskException):
                logger.exception(e)
                raise e
            __persisted_tasks.discard(task)

async def __persisted_task_exc_wrap(task: "asyncio.Task[T]") -> T:
    try:
        await task
    except Exception as e:
        raise exceptions.PersistedTaskException(e, task) from e

@overload
def _yield(key: K, value: V, yields: Literal['keys']) -> K:...
@overload
def _yield(key: K, value: V, yields: Literal['both']) -> Tuple[K, V]:...
def _yield(key: K, value: V, yields: Literal['keys', 'both']) -> Union[K, Tuple[K, V]]:
    if yields == 'both':
        return key, value
    elif yields == 'keys':
        return key
    else:
        raise ValueError(f"`yields` must be 'keys' or 'both'. You passed {yields}")

async def _yield_keys(iterable: AnyIterable[K]) -> AsyncIterator[K]:
    if isinstance(iterable, AsyncIterable):
        async for key in iterable:
            yield key
    elif isinstance(iterable, Iterable):
        for key in iterable:
            yield key
    else:
        raise TypeError(iterable)



import asyncio
import pytest

from a_sync import TaskMapping, create_task, exceptions

@pytest.mark.asyncio_cooperative
async def test_create_task():
    await create_task(coro=asyncio.sleep(0), name='test')

@pytest.mark.asyncio_cooperative
async def test_persistent_task():
    check = False
    async def task():
        await asyncio.sleep(1)
        nonlocal check
        check = True
    create_task(coro=task(), skip_gc_until_done=True)
    # there is no local reference to the newly created task. does it still complete? 
    await asyncio.sleep(2)
    assert check is True

@pytest.mark.asyncio_cooperative
async def test_pruning():
    async def task():
        return
    create_task(coro=task(), skip_gc_until_done=True)
    await asyncio.sleep(0)
    # previously, it failed here
    create_task(coro=task(), skip_gc_until_done=True)

@pytest.mark.asyncio_cooperative
async def test_task_mapping_init():
    tasks = TaskMapping(_coro_fn)
    assert tasks._coro_fn is _coro_fn
    assert tasks._coro_fn_kwargs == {}
    assert tasks._name == ""
    tasks = TaskMapping(_coro_fn, name='test', kwarg0=1, kwarg1=None)
    assert tasks._coro_fn_kwargs == {'kwarg0': 1, 'kwarg1': None}
    assert tasks._name == "test"

@pytest.mark.asyncio_cooperative
async def test_task_mapping():
    tasks = TaskMapping(_coro_fn)
    # does it return the correct type
    assert isinstance(tasks[0], asyncio.Task)
    # does it correctly return existing values
    assert tasks[1] is tasks[1]
    # does the task return the correct value
    assert await tasks[0] == "1"
    # can it do it again
    assert await tasks[0] == "1"
    
@pytest.mark.asyncio_cooperative
async def test_task_mapping_map_with_sync_iter():
    tasks = TaskMapping(_coro_fn)
    async for k, v in tasks.map(range(5)):
        # this shouldn't work since there is a mapping in progress
        with pytest.raises(exceptions.MappingNotEmptyError):
            async for k in tasks.map(range(5)):
                ...
        assert isinstance(k, int)
        assert isinstance(v, str)
    tasks = TaskMapping(_coro_fn)
    async for k in tasks.map(range(5), yields='keys'):
        assert isinstance(k, int)
    
@pytest.mark.asyncio_cooperative
async def test_task_mapping_map_with_async_iter():
    async def async_iter():
        for i in range(5):
            yield i
    tasks = TaskMapping(_coro_fn)
    async for k, v in tasks.map(async_iter()):
        assert isinstance(k, int)
        assert isinstance(v, str)
        # this shouldn't work since there is a mapping in progress
        with pytest.raises(exceptions.MappingNotEmptyError):
            async for k in tasks.map(async_iter()):
                ...
    tasks = TaskMapping(_coro_fn)
    async for k in tasks.map(async_iter(), yields='keys'):
        assert isinstance(k, int)

async def _coro_fn(i: int) -> str:
    i += 1
    return str(i) * i




from a_sync import aliases
from a_sync.base import ASyncGenericBase
from a_sync.decorator import a_sync
from a_sync.future import ASyncFuture, future  # type: ignore [attr-defined]
from a_sync.iter import ASyncIterable, ASyncIterator
from a_sync.modifiers.semaphores import apply_semaphore
from a_sync.primitives import *
from a_sync.singleton import ASyncGenericSingleton
from a_sync.task import TaskMapping as map
from a_sync.task import TaskMapping, create_task
from a_sync.utils import all, any, as_completed, as_yielded, gather

# I alias the aliases for your convenience.
# I prefer "aka" but its meaning is not intuitive when reading code so I created both aliases for you to choose from.
# NOTE: Overkill? Maybe.
aka = alias = aliases

# alias for backward-compatability, will be removed eventually, probably in 0.1.0
ASyncBase = ASyncGenericBase


__all__ = [
    "all",
    "any",
    "as_completed",
    "as_yielded",
    "create_task",
    "exhaust_iterator",
    "exhaust_iterators",
    "gather", 
    "map",
    "ASyncIterable",
    "ASyncIterator",
    "ASyncGenericSingleton",
    "TaskMapping", 
]




import asyncio

try:
    from tqdm.asyncio import tqdm_asyncio
except ImportError as e:
    class tqdm_asyncio:  # type: ignore [no-redef]
        def as_completed(*args, **kwargs):
            raise ImportError("You must have tqdm installed to use this feature")
        
from a_sync._typing import *
from a_sync.iter import ASyncIterator

@overload
def as_completed(fs: Iterable[Awaitable[T]], *, timeout: Optional[float] = None, return_exceptions: bool = False, aiter: Literal[False] = False, tqdm: bool = False, **tqdm_kwargs: Any) -> Iterator[Coroutine[None, None, T]]:
    ...
@overload
def as_completed(fs: Iterable[Awaitable[T]], *, timeout: Optional[float] = None, return_exceptions: bool = False, aiter: Literal[True] = True, tqdm: bool = False, **tqdm_kwargs: Any) -> ASyncIterator[T]:
    ...
@overload
def as_completed(fs: Mapping[K, Awaitable[V]], *, timeout: Optional[float] = None, return_exceptions: bool = False, aiter: Literal[False] = False, tqdm: bool = False, **tqdm_kwargs: Any) -> Iterator[Coroutine[None, None, Tuple[K, V]]]:
    ...
@overload
def as_completed(fs: Mapping[K, Awaitable[V]], *, timeout: Optional[float] = None, return_exceptions: bool = False, aiter: Literal[True] = True, tqdm: bool = False, **tqdm_kwargs: Any) -> ASyncIterator[Tuple[K, V]]:
    ...
def as_completed(fs, *, timeout: Optional[float] = None, return_exceptions: bool = False, aiter: bool = False, tqdm: bool = False, **tqdm_kwargs: Any):
    """
    Concurrently awaits a list of awaitable objects or mappings of awaitables and returns an iterator of results.

    This function extends Python's asyncio.as_completed, providing additional features for mixed use cases of individual awaitable objects and mappings of awaitables.

    Differences from asyncio.as_completed:
    - Uses type hints for use with static type checkers.
    - Supports either individual awaitables or a k:v mapping of awaitables.
    - Can be used as an async iterator which yields the result values. Example below.
    - Provides progress reporting using tqdm if 'tqdm' is set to True.

    Args:
        fs (Iterable[Awaitable[T] or Mapping[K, Awaitable[V]]]): The awaitables to await concurrently. It can be a list of individual awaitables or a mapping of awaitables.
        timeout (float, optional): The maximum time, in seconds, to wait for the completion of awaitables. Defaults to None (no timeout).
        return_exceptions (bool, optional): If True, exceptions are returned as results instead of raising them. Defaults to False.
        aiter (bool, optional): If True, returns an async iterator of results. Defaults to False.
        tqdm (bool, optional): If True, enables progress reporting using tqdm. Defaults to False.
        **tqdm_kwargs: Additional keyword arguments for tqdm if progress reporting is enabled.

    Returns:
        Iterator[Coroutine[None, None, T] or ASyncIterator[Tuple[K, V]]]: An iterator of results when awaiting individual awaitables or an async iterator when awaiting mappings.

    Examples:
        Awaiting individual awaitables:
        ```
        awaitables = [async_function1(), async_function2()]
        for coro in as_completed(awaitables):
            val = await coro
            ...
            
        async for val in as_completed(awaitables, aiter=True):
            ...
        ```
        
        Awaiting mappings of awaitables:
        ```
        mapping = {'key1': async_function1(), 'key2': async_function2()}
        
        for coro in as_completed(mapping):
            k, v = await coro
            ...
            
        async for k, v in as_completed(mapping, aiter=True):
            ...
        ```
    """
    if isinstance(fs, Mapping):
        return as_completed_mapping(fs, timeout=timeout, return_exceptions=return_exceptions, aiter=aiter, tqdm=tqdm, **tqdm_kwargs) 
    if return_exceptions:
        fs = [_exc_wrap(f) for f in fs]
    return (
        ASyncIterator.wrap(__yield_as_completed(fs, timeout=timeout, tqdm=tqdm, **tqdm_kwargs)) if aiter 
        else tqdm_asyncio.as_completed(fs, timeout=timeout, **tqdm_kwargs) if tqdm 
        else asyncio.as_completed(fs, timeout=timeout)
    )

@overload
def as_completed_mapping(mapping: Mapping[K, Awaitable[V]], *, timeout: Optional[float] = None, return_exceptions: bool = False, aiter: Literal[True] = True, tqdm: bool = False, **tqdm_kwargs: Any) -> ASyncIterator[Tuple[K, V]]:
    ...
@overload
def as_completed_mapping(mapping: Mapping[K, Awaitable[V]], *, timeout: Optional[float] = None, return_exceptions: bool = False, aiter: Literal[False] = False, tqdm: bool = False, **tqdm_kwargs: Any) -> Iterator[Coroutine[None, None, Tuple[K, V]]]:
    ...
def as_completed_mapping(mapping: Mapping[K, Awaitable[V]], *, timeout: Optional[float] = None, return_exceptions: bool = False, aiter: bool = False, tqdm: bool = False, **tqdm_kwargs: Any) -> Union[Iterator[Coroutine[None, None, Tuple[K, V]]], ASyncIterator[Tuple[K, V]]]:
    """
    Concurrently awaits a mapping of awaitable objects and returns an iterator or async iterator of results.

    This function is designed to await a mapping of awaitable objects, where each key-value pair represents a unique awaitable. It enables concurrent execution and gathers results into an iterator or an async iterator.

    Args:
        mapping (Mapping[K, Awaitable[V]]): A dictionary-like object where keys are of type K and values are awaitable objects of type V.
        timeout (float, optional): The maximum time, in seconds, to wait for the completion of awaitables. Defaults to None (no timeout).
        return_exceptions (bool, optional): If True, exceptions are returned as results instead of raising them. Defaults to False.
        aiter (bool, optional): If True, returns an async iterator of results. Defaults to False.
        tqdm (bool, optional): If True, enables progress reporting using tqdm. Defaults to False.
        **tqdm_kwargs: Additional keyword arguments for tqdm if progress reporting is enabled.

    Returns:
        Union[Iterator[Coroutine[None, None, Tuple[K, V]]] or ASyncIterator[Tuple[K, V]]]: An iterator of results or an async iterator when awaiting mappings.

    Example:
        ```
        mapping = {'key1': async_function1(), 'key2': async_function2()}
        
        for coro in as_completed_mapping(mapping):
            k, v = await coro
            ...
            
        async for k, v in as_completed_mapping(mapping, aiter=True):
            ...
        ```
    """
    return as_completed([__mapping_wrap(k, v, return_exceptions=return_exceptions) for k, v in mapping.items()], timeout=timeout, aiter=aiter, tqdm=tqdm, **tqdm_kwargs)

async def _exc_wrap(awaitable: Awaitable[T]) -> Union[T, Exception]:
    try:
        return await awaitable
    except Exception as e:
        return e
    
async def __yield_as_completed(futs: Iterable[Awaitable[T]], *, timeout: Optional[float] = None, return_exceptions: bool = False, tqdm: bool = False, **tqdm_kwargs: Any) -> AsyncIterator[T]:
    for fut in as_completed(futs, timeout=timeout, return_exceptions=return_exceptions, tqdm=tqdm, **tqdm_kwargs):
        yield await fut

@overload
async def __mapping_wrap(k: K, v: Awaitable[V], return_exceptions: Literal[True] = True) -> Union[V, Exception]:...
@overload
async def __mapping_wrap(k: K, v: Awaitable[V], return_exceptions: Literal[False] = False) -> V:...
async def __mapping_wrap(k: K, v: Awaitable[V], return_exceptions: bool = False) -> Union[V, Exception]:
    try:
        return k, await v
    except Exception as e:
        if return_exceptions:
            return k, e
        raise e
    


# mypy: disable-error-code=valid-type
import functools

from a_sync._typing import *
from a_sync.config import user_set_default_modifiers, null_modifiers
from a_sync.modifiers import cache, limiter, semaphores

valid_modifiers = [key for key in ModifierKwargs.__annotations__ if not key.startswith('_') and not key.endswith('_')]

class ModifierManager(Dict[str, Any]):
    default: DefaultMode
    cache_type: CacheType
    cache_typed: bool
    ram_cache_maxsize: Optional[int]
    ram_cache_ttl: Optional[int]
    runs_per_minute: Optional[int]
    semaphore: SemaphoreSpec
    # sync modifiers
    executor: Executor

    def __init__(self, modifiers: ModifierKwargs) -> None:
        for key in modifiers.keys():
            if key not in valid_modifiers:
                raise ValueError(f"'{key}' is not a supported modifier.")
        self._modifiers = modifiers
    def __repr__(self) -> str:
        return str(self._modifiers)
    def __getattribute__(self, modifier_key: str) -> Any:
        if modifier_key not in valid_modifiers:
            return super().__getattribute__(modifier_key)
        return self[modifier_key] if modifier_key in self else user_defaults[modifier_key]

    
    @property
    def use_limiter(self) -> bool:
        return self.runs_per_minute != nulls.runs_per_minute
    @property
    def use_semaphore(self) -> bool:
        return self.semaphore != nulls.semaphore
    @property
    def use_cache(self) -> bool:
        return any([
            self.cache_type != nulls.cache_type,
            self.ram_cache_maxsize != nulls.ram_cache_maxsize,
            self.ram_cache_ttl != nulls.ram_cache_ttl,
            self.cache_typed != nulls.cache_typed,
        ])
    
    def apply_async_modifiers(self, coro_fn: CoroFn[P, T]) -> CoroFn[P, T]:
        # NOTE: THESE STACK IN REVERSE ORDER
        if self.use_limiter:
            coro_fn = limiter.apply_rate_limit(coro_fn, self.runs_per_minute)
        if self.use_semaphore:
            coro_fn = semaphores.apply_semaphore(coro_fn, self.semaphore)
        if self.use_cache:
            coro_fn = cache.apply_async_cache(
                coro_fn,
                cache_type=self.cache_type or 'memory',
                cache_typed=self.cache_typed,
                ram_cache_maxsize=self.ram_cache_maxsize,
                ram_cache_ttl=self.ram_cache_ttl
            )
        return coro_fn
    
    def apply_sync_modifiers(self, function: SyncFn[P, T]) -> SyncFn[P, T]:
        @functools.wraps(function)
        def sync_modifier_wrap(*args: P.args, **kwargs: P.kwargs) -> T:
            return function(*args, **kwargs)
        # NOTE There are no sync modifiers at this time but they will be added here for my convenience.
        return sync_modifier_wrap
    
    # Dictionary api
    def keys(self) -> KeysView[str]:  # type: ignore [override]
        return self._modifiers.keys()
    def values(self) -> ValuesView[Any]:  # type: ignore [override]
        return self._modifiers.values()
    def items(self) -> ItemsView[str, Any]:  # type: ignore [override]
        return self._modifiers.items()
    def __contains__(self, key: str) -> bool:  # type: ignore [override]
        return key in self._modifiers
    def __iter__(self) -> Iterator[str]:
        return self._modifiers.__iter__()
    def __len__(self) -> int:
        return len(self._modifiers)
    def __getitem__(self, modifier_key: str):
        return self._modifiers[modifier_key]  # type: ignore [literal-required]

nulls = ModifierManager(null_modifiers)
user_defaults = ModifierManager(user_set_default_modifiers)



import asyncio
import time

import pytest

from a_sync.primitives.executor import (AsyncProcessPoolExecutor,
                                        ProcessPoolExecutor)


def do_work(i, kwarg=None):
    time.sleep(i)
    if kwarg:
        assert kwarg == i
    
def test_executor():
    executor = ProcessPoolExecutor(1)
    assert isinstance(executor, AsyncProcessPoolExecutor)
    coro = executor.run(do_work, 3)
    # TODO: make `submit` return asyncio.Future
    asyncio.get_event_loop().run_until_complete(coro)
    fut = executor.submit(do_work, 3)
    assert isinstance(fut, asyncio.Future), fut.__dict__
    asyncio.get_event_loop().run_until_complete(fut)
    
    # asyncio implementation cant handle kwargs :(
    with pytest.raises(TypeError):
        asyncio.get_event_loop().run_until_complete(
            asyncio.get_event_loop().run_in_executor(executor, do_work, 3, kwarg=3)
        )
        
    # but our clean implementation can :)
    fut = executor.submit(do_work, 3, kwarg=3)
    
    asyncio.get_event_loop().run_until_complete(
        executor.run(do_work, 3, kwarg=3)
    )
    asyncio.get_event_loop().run_until_complete(
        fut
    )
    